'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.to5BitArray = to5BitArray;
exports.from5BitArray = from5BitArray;
exports.encode5BitArray = encode5BitArray;
exports.encode = encode;
exports.decodeTo5BitArray = decodeTo5BitArray;
exports.decode = decode;

var _bitConverter = require('./bit-converter');

var _encoding = require('./encoding');

// Minimum char code that could be present in the encoded message
var MIN_CHAR_CODE = 33;
// Maximum char code that could be present in the encoded message
var MAX_CHAR_CODE = 126;
// Maximum encoded message length
var MAX_ENC_LENGTH = 90;

/**
 * Converts a Uint8Array into a Uint8Array variant, in which each element
 * encodes 5 bits of the original byte array.
 *
 * @param {Uint8Array} src
 *   Input to convert
 * @param {?Uint8Array} outBuffer
 *   Optional output buffer. If specified, the 5-bit sequence will be written there;
 *   if not specified, the output buffer will be created from scratch. The length
 *   of `outBuffer` is not checked.
 * @returns {Uint8Array}
 *   Output buffer with a 5-bit sequence
 *
 * @api public
 */
function to5BitArray(src, dst) {
  var len = Math.ceil(src.length * 8 / 5);
  var realDst = dst || new Uint8Array(len);

  return (0, _bitConverter.toBits)(src, 5, realDst);
}

function from5BitArray(src, dst) {
  var len = Math.floor(src.length * 5 / 8);
  var realDst = dst || new Uint8Array(len);

  return (0, _bitConverter.fromBits)(src, 5, realDst);
}

/**
 * Encodes binary data into Bech32 encoding.
 *
 * Ordinarily, you may want to use [`encode`](#encode) because it converts
 * binary data to an array of 5-bit integers automatically.
 *
 * @param {string} prefix
 *   human-readable prefix to place at the beginning of the encoding
 * @param {Uint8Array} data
 *   array of 5-bit integers with data to encode
 * @returns {string}
 *   Bech32 encoding of data in the form `<prefix>1<base32 of data><checksum>`
 *
 * @api public
 */
function encode5BitArray(prefix, data) {
  // 1. Allocate buffer for all operations
  var len = 2 * prefix.length + 1 + // expanded prefix
  data.length + // five-bit data encoding
  _encoding.CHECKSUM_LENGTH; // checksum

  if (len - prefix.length > MAX_ENC_LENGTH) {
    throw new Error('Message to be produced is too long (max ' + MAX_ENC_LENGTH + ' supported)');
  }

  for (var i = 0; i < prefix.length; i++) {
    var ord = prefix.charCodeAt(i);
    if (ord < MIN_CHAR_CODE || ord > MAX_CHAR_CODE) {
      throw new TypeError('Invalid char in prefix: ' + ord + '; should be in ASCII range ' + MIN_CHAR_CODE + '-' + MAX_CHAR_CODE);
    }
  }

  var buffer = new Uint8Array(len);

  // 2. Expand the human-readable prefix into the beginning of the buffer
  (0, _encoding.expandPrefix)(prefix, buffer.subarray(0, 2 * prefix.length + 1));

  // 3. Copy `data` into the output
  var dataBuffer = buffer.subarray(2 * prefix.length + 1, buffer.length - _encoding.CHECKSUM_LENGTH);
  dataBuffer.set(data);

  // 4. Create the checksum
  (0, _encoding.createChecksum)(buffer);

  // 5. Convert into string
  var encoded = (0, _encoding.encode)(buffer.subarray(2 * prefix.length + 1));
  return prefix + '1' + encoded;
}

/**
 * Encodes binary data into Bech32 encoding.
 *
 * @param {string} prefix
 *   human-readable prefix to place at the beginning of the encoding
 * @param {Uint8Array} data
 *   binary data to encode
 * @returns {string}
 *   Bech32 encoding of data in the form `<prefix>1<base32 of data><checksum>`
 *
 * @api public
 */
function encode(prefix, data) {
  return encode5BitArray(prefix, to5BitArray(data));
}

/**
 * Decodes data from Bech32 encoding into an array of 5-bit integers.
 *
 * Ordinarily, you may want to use [`decode`](#decode) because it automatically
 * converts the array of 5-bit integers into an ordinary `Uint8Array`.
 *
 * @param {string} message
 *   Bech32-encoded message
 * @returns {Object}
 *   Decoded object with `prefix` and `data` fields, which contain the human-readable
 *   prefix and the array of 5-bit integers respectively.
 *
 * @api public
 */
function decodeTo5BitArray(message) {
  // Check preconditions

  // 1. Message length
  if (message.length > MAX_ENC_LENGTH) {
    throw new TypeError('Message too long; max ' + MAX_ENC_LENGTH + ' expected');
  }

  // 2. Mixed case
  var hasLowerCase = false;
  var hasUpperCase = false;
  for (var i = 0; i < message.length; i++) {
    var ord = message.charCodeAt(i);

    // 3. Allowed chars in the encoding
    if (ord < MIN_CHAR_CODE || ord > MAX_CHAR_CODE) {
      throw new TypeError('Invalid char in message: ' + ord + '; should be in ASCII range ' + MIN_CHAR_CODE + '-' + MAX_CHAR_CODE);
    }
    hasLowerCase = hasLowerCase || ord >= 65 && ord <= 90;
    hasUpperCase = hasUpperCase || ord >= 97 && ord <= 122;
  }
  if (hasLowerCase && hasUpperCase) {
    throw new TypeError('Mixed-case message');
  }

  var lowerCaseMsg = message.toLowerCase();

  // 4. Existence of the separator char
  var sepIdx = lowerCaseMsg.lastIndexOf('1');
  if (sepIdx < 0) {
    throw new Error('No separator char ("1") found');
  }

  // 5. Placing of the separator char in the message
  if (sepIdx > message.length - _encoding.CHECKSUM_LENGTH - 1) {
    throw new Error('Data part of the message too short (at least ' + _encoding.CHECKSUM_LENGTH + ' chars expected)');
  }

  var prefix = lowerCaseMsg.substring(0, sepIdx);

  // Checked within `decodeWithPrefix`:
  // 6. Invalid chars in the data part of the message
  var bitArray = (0, _encoding.decodeWithPrefix)(prefix, lowerCaseMsg.substring(sepIdx + 1));

  // 7. Checksum
  if (!(0, _encoding.verifyChecksum)(bitArray)) {
    throw new Error('Invalid checksum');
  }

  return {
    prefix: prefix,
    // Strip off the prefix from the front and the checksum from the end
    data: bitArray.subarray(2 * prefix.length + 1, bitArray.length - _encoding.CHECKSUM_LENGTH)
  };
}

/**
 * Decodes data from Bech32 encoding into an array of 5-bit integers.
 *
 * @param {string} message
 *   Bech32-encoded message
 * @returns {Object}
 *   Decoded object with `prefix` and `data` fields, which contain the human-readable
 *   prefix and the decoded binary data respectively.
 *
 * @api public
 */
function decode(message) {
  var _decodeTo5BitArray = decodeTo5BitArray(message),
      prefix = _decodeTo5BitArray.prefix,
      bitArray = _decodeTo5BitArray.data;

  return { prefix: prefix, data: from5BitArray(bitArray) };
}